/*
 * linux/drivers/pcmcia/pxa2xx_colibri.c
 *
 * CompactFlash support for the Toradex Colibri XScale modules
 *
 * Copyright (C) 2008 Noser Engineering
 *
 * 2008-04-07: Marcel Ziswiler <marcel.ziswiler@noser.com>
 *             initial version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>

#include <asm/mach-types.h>

#include <linux/delay.h>
#include <linux/platform_device.h>

#include <mach/gpio.h>
/* TODO: move to board specific file */
#include <mach/mfp-pxa320.h>

#include "soc_common.h"

/* TODO: move to board specific file */
static int gpio_ready	= mfp_to_gpio(MFP_PIN_GPIO29);
static int gpio_bvd1	= mfp_to_gpio(MFP_PIN_GPIO53);
static int gpio_power	= mfp_to_gpio(MFP_PIN_GPIO57);
static int gpio_reset	= mfp_to_gpio(MFP_PIN_GPIO77);
static int gpio_bvd2	= mfp_to_gpio(MFP_PIN_GPIO79);
static int gpio_detect	= mfp_to_gpio(MFP_PIN_GPIO81);

static struct pcmcia_irqs irqs[] = {
	{ 0, 0, "CompactFlash card detect" },
};

static int colibri_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
	int err = 0;

	/* Initialise GPIOs */
	err = gpio_request(gpio_ready, "CompactFlash card ready");
	if (err)
		goto err_request_ready;
	gpio_direction_input(gpio_ready);
	err = gpio_request(gpio_bvd1, "CompactFlash card BVD1");
	if (err)
		goto err_request_bvd1;
	gpio_direction_input(gpio_bvd1);
	err = gpio_request(gpio_power, "CompactFlash power enable");
	if (err)
		goto err_request_power;
	gpio_direction_output(gpio_power, true); /* active low */
	err = gpio_request(gpio_reset, "CompactFlash reset");
	if (err)
		goto err_request_reset;
	gpio_direction_output(gpio_reset, false);
	err = gpio_request(gpio_detect, "CompactFlash card detect");
	if (err)
		goto err_request_detect;
	gpio_direction_input(gpio_detect);
	err = gpio_request(gpio_bvd2, "CompactFlash card BVD2");
	if (err)
		goto err_request_bvd2;
	gpio_direction_input(gpio_bvd2);

	/* Switch power on (active low) */
	gpio_set_value(gpio_power, false);

	/* Reset PCMCIA controller */
	gpio_set_value(gpio_reset, true);
	udelay(500);
	gpio_set_value(gpio_reset, false);

	/* Set interrupts */
	skt->socket.pci_irq = gpio_to_irq(gpio_ready);
	irqs[0].irq = gpio_to_irq(gpio_detect);
	err = soc_pcmcia_request_irqs(skt, irqs, ARRAY_SIZE(irqs));
	if (err)
		goto err_request_interrupts;

	return 0;

err_request_interrupts:
	gpio_free(gpio_bvd2);
err_request_bvd2:
	gpio_free(gpio_detect);
err_request_detect:
	gpio_free(gpio_reset);
err_request_reset:
	gpio_free(gpio_power);
err_request_power:
	gpio_free(gpio_bvd1);
err_request_bvd1:
	gpio_free(gpio_ready);
err_request_ready:
	return err;
}

static void colibri_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	/* Free interrupts */
	soc_pcmcia_free_irqs(skt, irqs, ARRAY_SIZE(irqs));

	/* Free GPIOs */
	gpio_free(gpio_bvd2);
	gpio_free(gpio_detect);
	gpio_free(gpio_reset);
	gpio_free(gpio_bvd1);
	gpio_free(gpio_ready);

	/* Switch power off (active low) */
	gpio_set_value(gpio_power, true);
	gpio_free(gpio_power);
}

static void colibri_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
					struct pcmcia_state *state)
{
	state->detect = gpio_get_value(gpio_detect)?1:0;
	state->ready  = gpio_get_value(gpio_ready)?1:0;
	state->bvd1  = gpio_get_value(gpio_bvd1)?1:0;
	state->bvd2  = gpio_get_value(gpio_bvd2)?1:0;
	state->vs_3v  = 1;	/* Only 3.3 V supported */
	state->vs_Xv  = 0;
	state->wrprot = 0;	/* Not available, r/w all the time */
}

static int colibri_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
					   const socket_state_t *state)
{
	unsigned long flags;

	local_irq_save(flags);

	/* Configure Vcc and Vpp */
	if (state->Vcc == 0 && state->Vpp == 0)
		gpio_set_value(gpio_power, true);	/* active low */
	else if (state->Vcc == 33 && state->Vpp < 50)
		gpio_set_value(gpio_power, false);	/* active low */
	else {
		printk(KERN_ERR "%s(): unsupported Vcc %u Vpp %u combination\n",
		       __func__, state->Vcc, state->Vpp);
		return -1;
	}

	/* Reset if requested */
	gpio_set_value(gpio_reset, (state->flags & SS_RESET)?1:0);

	local_irq_restore(flags);
	udelay(200);

	return 0;
}

static void colibri_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
}

static void colibri_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
}

static struct pcmcia_low_level colibri_pcmcia_ops = {
	.owner			= THIS_MODULE,

	.first			= 0,
	.nr			= 1,

	.hw_init		= colibri_pcmcia_hw_init,
	.hw_shutdown		= colibri_pcmcia_hw_shutdown,

	.socket_state		= colibri_pcmcia_socket_state,
	.configure_socket	= colibri_pcmcia_configure_socket,

	.socket_init		= colibri_pcmcia_socket_init,
	.socket_suspend		= colibri_pcmcia_socket_suspend,
};

static struct platform_device *colibri_pcmcia_device;

static int __init colibri_pcmcia_init(void)
{
	int ret;

	if (!machine_is_colibri320())
		return -ENODEV;

	colibri_pcmcia_device = platform_device_alloc("pxa2xx-pcmcia", -1);
	if (!colibri_pcmcia_device)
		return -ENOMEM;

	ret = platform_device_add_data(colibri_pcmcia_device,
				       &colibri_pcmcia_ops,
			sizeof(colibri_pcmcia_ops));

	printk(KERN_INFO "Registering Colibri CompactFlash interface.\n");
	if (!ret)
		ret = platform_device_add(colibri_pcmcia_device);

	if (ret)
		platform_device_put(colibri_pcmcia_device);

	return ret;
}

static void __exit colibri_pcmcia_exit(void)
{
	platform_device_unregister(colibri_pcmcia_device);
}

fs_initcall(colibri_pcmcia_init);
module_exit(colibri_pcmcia_exit);

MODULE_ALIAS("platform:pxa2xx-pcmcia");
MODULE_AUTHOR("Marcel Ziswiler <marcel.ziswiler@noser.com>");
MODULE_DESCRIPTION("Colibri CompactFlash driver");
MODULE_LICENSE("GPL");
