/*
 * SoC audio driver for Toradex Limestone
 *
 * Copyright (C) 2010-2011 Noser Engineering
 *
 * 2010-10-02: Marcel Ziswiler <marcel.ziswiler@noser.com>
 *             initial version
 *
 * Copied from tosa.c:
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Copyright 2005 Openedhand Ltd.
 *
 * Authors: Liam Girdwood <lrg@slimlogic.co.uk>
 *          Richard Purdie <richard@openedhand.com>
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/gpio.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/mach-types.h>
#include <mach/audio.h>

#include "../codecs/wm9713.h"
#include "pxa2xx-pcm.h"
#include "pxa2xx-ac97.h"

static const struct snd_soc_dapm_widget limestone_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headset headphone", NULL),
	SND_SOC_DAPM_MIC("Headset MIC", NULL),
	SND_SOC_DAPM_MIC("MIC", NULL),
	SND_SOC_DAPM_SPK("Speaker", NULL),
};

/* Currently supported audio map */
static const struct snd_soc_dapm_route limestone_audio_map[] = {
	/* Headset jack X61 (HPL/R) */
	{ "Headset headphone", NULL, "HPL" },
	{ "Headset headphone", NULL, "HPR" },

	/* Headset jack X61 (MIC1) */
	{ "Headset MIC", NULL, "MIC1" },

	/* Microphone X1 (MICCM, MIC2A) */
	{ "MIC", NULL, "MIC2A" },

	/* Speaker X4 (SPKL/R) */
	{ "Speaker", NULL, "SPKL" },
	{ "Speaker", NULL, "SPKR" },
};

static int limestone_wm9713l_init(struct snd_soc_codec *codec)
{
	int err;

	/* add Limestone specific widgets */
	err = snd_soc_dapm_new_controls(codec, limestone_dapm_widgets,
					ARRAY_SIZE(limestone_dapm_widgets));
	if (err)
		return err;

	/* set up Limestone specific audio path audio_map */
	err = snd_soc_dapm_add_routes(codec, limestone_audio_map,
				      ARRAY_SIZE(limestone_audio_map));
	if (err)
		return err;

	/* connected pins */
	snd_soc_dapm_enable_pin(codec, "MIC1");
	snd_soc_dapm_enable_pin(codec, "MIC2A");
	snd_soc_dapm_enable_pin(codec, "HPL");
	snd_soc_dapm_enable_pin(codec, "HPR");
	snd_soc_dapm_enable_pin(codec, "SPKL");
	snd_soc_dapm_enable_pin(codec, "SPKR");

	/* not connected pins */
	snd_soc_dapm_nc_pin(codec, "MIC2B");
	snd_soc_dapm_nc_pin(codec, "LINEL");
	snd_soc_dapm_nc_pin(codec, "LINER");
	snd_soc_dapm_nc_pin(codec, "MONOIN");
	snd_soc_dapm_nc_pin(codec, "PCBEEP");
	snd_soc_dapm_nc_pin(codec, "MONO");
	snd_soc_dapm_nc_pin(codec, "OUT3");
	snd_soc_dapm_nc_pin(codec, "OUT4");

	err = snd_soc_dapm_sync(codec);
	if (err)
		return err;

	return 0;
}

static struct snd_soc_dai_link limestone_dai[] = {
	{
		.name = "AC97 HiFi",
		.stream_name = "AC97 HiFi",
		.cpu_dai = &pxa_ac97_dai[PXA2XX_DAI_AC97_HIFI],
		.codec_dai = &wm9713_dai[WM9713_DAI_AC97_HIFI],
		.init = limestone_wm9713l_init,
	},
	{
		.name = "AC97 Aux",
		.stream_name = "AC97 Aux",
		.cpu_dai = &pxa_ac97_dai[PXA2XX_DAI_AC97_AUX],
		.codec_dai = &wm9713_dai[WM9713_DAI_AC97_AUX],
	},
};

static struct snd_soc_card limestone = {
	.name = "Toradex Limestone",
	.platform = &pxa2xx_soc_platform,
	.dai_link = limestone_dai,
	.num_links = ARRAY_SIZE(limestone_dai),
};

static struct snd_soc_device limestone_snd_devdata = {
	.card = &limestone,
	.codec_dev = &soc_codec_dev_wm9713,
};

static struct platform_device *limestone_snd_device;

static int __init limestone_init(void)
{
	int ret;

	if (!(machine_is_limestone()))
		return -ENODEV;

	limestone_snd_device = platform_device_alloc("soc-audio", -1);
	if (!limestone_snd_device)
		return -ENOMEM;

	platform_set_drvdata(limestone_snd_device, &limestone_snd_devdata);
	limestone_snd_devdata.dev = &limestone_snd_device->dev;
	ret = platform_device_add(limestone_snd_device);

	if (!ret)
		return 0;

	/* Fail gracefully */
	platform_device_put(limestone_snd_device);

	return ret;
}

static void __exit limestone_exit(void)
{
	platform_device_unregister(limestone_snd_device);
}

module_init(limestone_init);
module_exit(limestone_exit);

/* Module information */
MODULE_AUTHOR("Marcel Ziswiler");
MODULE_DESCRIPTION("ALSA SoC WM9713L on Toradex Limestone");
MODULE_LICENSE("GPL");
